# icon.rb: Supporting library for icon artists, icon generation and
# check utilities
#
# Copyright (C) 2008--2013 Martin Sourada
#
# This script is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This script is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this script; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

require "icon-artist/theme-settings"
require "fileutils"
require "rexml/document"

#include REXML
include FileUtils

module IconArtist
  INKSCAPE = '/usr/bin/inkscape'

  class Icon

    private

    def Icon.get_templatedir
      config = File.expand_path("~/.config/user-dirs.dirs")
      dir = "~/Templates"
      if File.exist?(config)
        f = File.open(config,"r")
          dirs = IO.readlines("#{config}")
          dirs.each do |line|
            if line.include?("XDG_TEMPLATES_DIR=")
              dir = line.gsub("XDG_TEMPLATES_DIR=","").delete('"').rstrip.gsub("$HOME","~")
            end
          end
        f.close
      end
      return dir
    end

    def Icon.check_context(context)
      if !THEME_CONTEXTS.include? context
        puts "Error: Unknown icon context '#{context}'. \nContext must be one of: "
        p THEME_CONTEXTS
        return FALSE
      end

      return TRUE
    end

    def Icon.check_document_metadata(svg, filename, multiple_icons = false)
      # Check for icon name
      messages = {:errors => [], :warnings => []}
      unless multiple_icons
        element = svg.root.elements['//cc:Work/dc:title']
        if (element)
          title = element.text
          if !title
            msg = "Error: Missing document metadata title (icon name)."
            messages[:errors] += [msg]
            puts "#{msg}"
          else
            iconname = File.basename(filename, ".svg")
            if title != iconname
              msg = "Warning: Icon name in document metadata differs from file name.\n"
              msg += "Icon name in document metadata is: #{title}\n"
              msg += "Icon file name (without extension) is: #{iconname}"
              messages[:warnings] += [msg]
              puts "#{msg}"
            end
          end
        else
          msg = "Error: Missing document metadata title (icon name)."
          messages[:errors] += [msg]
          puts "#{msg}"
        end
      end

      # Check for icon context
      element = svg.root.elements['//dc:description']
      if element
        context = element.text
        if !context
          msg =  "Error: Missing document metadata description (icon context)."
          messages[:errors] += [msg]
          puts "#{msg}"
        else
          if !check_context(context)
            messages[:errors] += ["Error: Unknown icon context '#{context}'."]
          end
        end
      else
        msg = "Error: Missing document metadata description (icon context)."
        messages[:errors] += [msg]
        puts "#{msg}"
      end

      return messages

    end

    def Icon.check_layer_plate(svg)
      # Check if plate layer is present and hidden
      messages = {:errors => [], :warnings => []}
      plate_present = FALSE
      svg.root.each_element("//g[@inkscape:label='plate']") do |plate|
        plate_present = TRUE
        style = plate.attributes['style']
        if style != "display:none"
          msg = "Error: Plate layer is not hidden."
          messages[:errors] += [msg]
          puts "#{msg}"
        end
      end

      if !plate_present
        msg = "Error: Plate layer is missing."
        messages[:errors] += [msg]
        puts "#{msg}"
      end

      # Check for extra sizes/duplicates in plate layer
      plate_sizes = ['scalable']
      svg.root.each_element("//g[@inkscape:label='plate']/rect") do |icon|
        size = "#{icon.attributes['width']}x#{icon.attributes['height']}"
        if THEME_SIZES.include? "#{size}"
          if plate_sizes.include? "#{size}"
            msg = "Error: Size #{size} detected multiple times."
            messages[:errors] += [msg]
            puts "#{msg}"
          else
            plate_sizes += ["#{size}"]
          end
        else
          msg = "Warrning: Unknown icon size present in plate layer: #{size}."
          messages[:warnings] += [msg]
          puts "#{msg}"
        end
      end

      # Check for missing sizes in plate layer
      THEME_SIZES.each do |icon|
        if !plate_sizes.include? icon
          msg = "Warning: Size #{icon} missing from plate layer."
          messages[:warnings] += [msg]
          puts "#{msg}"
        end
      end

      # Check for scalable icon canvas
      scalable = svg.root.elements["//g[@inkscape:label='plate']/rect[@inkscape:label='scalable']"]
      if !scalable
        msg = "Error: Missing element with inkscape:label='scalable' in plate layer."
        messages[:errors] += [msg]
        puts "#{msg}"
      end

      return messages
    end

    def Icon.check_layer_artwork(svg, multiple_icons = false)
      # Check if artwork layer is present
      messages = {:errors => [], :warnings => []}
      if multiple_icons
        svg.root.elements["/svg"].each_element_with_attribute(
          "inkscape:groupmode", "layer", 0, "g") do |layer|
          layer_name = layer.attribute("inkscape:label").value
          if (layer_name.match("artwork-+"))
            scalable = layer.elements["g[@inkscape:label='scalable']"]
            if !scalable
              msg = "Error: Missing object group with inkscape:label='scalable' in '#{layer_name}' layer."
              messages[:errors] += [msg]
              puts "#{msg}"
            end
          end
        end
      else
        artwork = svg.root.elements["//g[@inkscape:label='artwork']"]
        if !artwork
          msg = "Error: Artwork layer is missing."
          messages[:errors] += [msg]
          puts "#{msg}"
        end

        # Check for scalable icon artwork
        scalable = svg.root.elements["//g[@inkscape:label='artwork']/g[@inkscape:label='scalable']"]
        if !scalable
          msg = "Error: Missing object group with inkscape:label='scalable' in artwork layer."
          messages[:errors] += [msg]
          puts "#{msg}"
        end
      end

      return messages
    end

    def Icon.detect_multiple(svg)
      multiple = false
      icon_names = []
      svg.root.elements["/svg"].each_element_with_attribute(
        "inkscape:groupmode", "layer", 0, "g") do |layer|
        layer_name = layer.attribute("inkscape:label").value
        if (layer_name.match("artwork-+"))
          multiple = true
          icon_names += [layer_name.split("artwork-")[1]]
        end
      end

      if multiple
        return icon_names
      else
        return nil
      end
    end

    def export_pngs(export_dir,sizes)
      return nil if @multiple_icons
      export_sizes = []

      @svg.root.each_element("//g[@inkscape:label='plate']/rect") do |icon|
        size = "#{icon.attributes['width']}x#{icon.attributes['height']}"
        if THEME_SIZES.include?(size) and sizes.include?(size)
          puts "Exporting #{size} size..."
          icon_dir = "#{export_dir}/#{size}/#{@context}"
          FileUtils.mkdir_p(icon_dir) unless File.exist?(icon_dir)
          if system "#{INKSCAPE} -i #{icon.attributes['id']} -e #{icon_dir}/#{@iconname}.png #{@filename} > /dev/null 2>&1"
            export_sizes += ["#{size}"]
          end
        end
      end

      return export_sizes
    end

    def export_scalable(export_dir)
      return nil if @multiple_icons
      puts "Making scalable size..."
      sizes = []

      scalable = REXML::Document.new(File.new("#{@filename}", 'r'))

      plate = scalable.root.elements["//g[@inkscape:label='plate']/rect[@inkscape:label='scalable']"]
      x = plate.attributes['x']
      y = plate.attributes['y']

      artwork = scalable.root.elements["//g[@inkscape:label='artwork']/g[@inkscape:label='scalable']"]
      #XXX Need update to work with icons that already has transform
      translate = "translate(-#{x},-#{y})"
      artwork.add_attribute(REXML::Attribute.new('transform', translate))

      canvas = scalable.root.elements["/svg"]
      canvas.add_attribute(REXML::Attribute.new('width', '48'))
      canvas.add_attribute(REXML::Attribute.new('height', '48'))

      plate = scalable.root.elements["//g[@inkscape:label='plate']"]
      plate.elements.delete_all('')

      scalable.root.each_element("//g[@inkscape:label='artwork']/*") do |obj|
        label = obj.attributes['inkscape:label']
        if (label != 'scalable')
          obj.parent.delete_element(obj)
        end
      end

      icon_dir = "#{export_dir}/scalable/#{@context}"
      out ="#{icon_dir}/#{@iconname}.svg"
      FileUtils.mkdir_p(icon_dir) unless File.exist?(icon_dir)

      f = File.open(out,"w")
      scalable.write(f, -1, false)
      f.close

      if system "#{INKSCAPE} --vacuum-defs -l #{out} -f #{out} > /dev/null 2>&1"
        sizes += ['scalable']
      end

      return sizes
    end

    def initialize_symlinks_icon(icon, iconname)
      symlinks_def = []
      symlinks_opt = []

      if icon != nil
        icon.each_element('link') do |link|
          symlinks_def = symlinks_def + [link.text]
        end
      end

      @symlinks[iconname] = {:default => symlinks_def, :optional => symlinks_opt}
    end

    def initialize_symlinks
      icon_mapping = "/usr/share/icon-naming-utils/legacy-icon-mapping.xml"
      icon_map = REXML::Document.new(File.new(icon_mapping, 'r'))


      @symlinks = {}
      if @multiple_icons
        @icons.each do |iconname|
          icon = icon_map.root.elements["//context[@dir='#{@context}']/icon[@name='#{iconname}']"]
          initialize_symlinks_icon(icon, iconname)
          @symlinks_initialized = true
        end
      else
          icon = icon_map.root.elements["//context[@dir='#{@context}']/icon[@name='#{@iconname}']"]
        initialize_symlinks_icon(icon, @iconname)
        @symlinks_initialized = true
      end
    end

    public

    # Open new icon from set file
    #
    # filename::
    #  path to a file that contains the icon
    #
    # options::
    #    :skip_check => don't check if the icon adheres one-canvas guidelines
    def initialize(filename,opts={})
      @symlinks_initialized = false
      @filename = filename
      @svg = REXML::Document.new(File.new("#{filename}", 'r'))

      @icons = Icon.detect_multiple(@svg)
      @multiple_icons = (@icons != nil)

      unless opts[:skip_check]
        messages = Icon.check_document_metadata(@svg, @filename, @multiple_icons)
        if messages[:errors].length > 0
          raise "Failed. Check above messages for more info."
        end

        msg = Icon.check_layer_plate(@svg)
        messages[:warnings] += msg[:warnings]
        messages[:errors] += msg[:errors]
        if messages[:errors].length > 0
          raise "Failed. Check above messages for more info."
        end

        msg = Icon.check_layer_artwork(@svg, @multiple_icons)
        messages[:warnings] += msg[:warnings]
        messages[:errors] += msg[:errors]
        if messages[:errors].length > 0
          raise "Failed. Check above messages for more info."
        end
      else
        messages = {:errors => [], :warnings => []}
      end

      @iconname = @svg.root.elements['//cc:Work/dc:title'].text
      @icons = [@iconname] unless @icons
      @context = @svg.root.elements['//dc:description'].text

      puts "Icon #{@filename} loaded successfully."
      if messages[:warnings].length > 0
        if messages[:warnings].length == 1
          puts "There was #{messages[:warnings].length} warning."
        else
          puts "There were #{messages[:warnings].length} warnings."
        end
      end
    end

    # Create new icon with set name and context using selected template
    #
    # iconname::
    #  name of the icon, preferably following freedesktop.org
    #  standards, to be created
    #
    # context::
    #  context of the icon, see THEME_CONTEXTS constant for more info
    #
    # template::
    #  select custom template, either template name (without .svg
    #  extension) that is located in your XDG User Dirs Templates dir or in
    #  /usr/share/icon-artist/templates/#{template}.svg, or a filename of a
    #  custom template (with the .svg extension)
    def self.create(iconname, context, template="default", dir = './')
      userdir_templates = Icon.get_templatedir
      @template = File.expand_path("#{template}")
      if !File.exist?(@template)
        @template = File.expand_path("#{userdir_templates}/#{template}.svg")
        if !File.exist?(@template)
          @template = "#{ARTIST_SHAREDIR}/templates/#{template}.svg"
          if !File.exist?(@template)
            raise "Could not find either '#{File.expand_path(template)}', '#{File.expand_path(userdir_templates)}/#{template}.svg' or '#{ARTIST_SHAREDIR}/templates/#{template}.svg'."
          end
        end
      end

      @iconname=iconname
      @context=context
      @filename=File.expand_path(dir)+"/#{@iconname}.svg"
      @multiple_icons = false
      @icons = [@iconname]

      if File.exist?(@filename)
        raise "Selected icon (#{@filename}) already exists. \nPlease set another icon name."
      end

      if !(THEME_CONTEXTS.include?(@context))
        out = "Context must be one of: "
        THEME_CONTEXTS.each {|el| out = out + "#{el}, "}
        out = (out + "\n").gsub(", \n", ".\n")
        raise out
      end

      @svg = REXML::Document.new(File.new("#{@template}", 'r'))

      @svg.root.attributes.delete_all("inkscape:export-filename")
      @svg.root.attributes.delete_all("sodipodi:docname")

      element = @svg.root.elements["metadata/rdf:RDF/cc:Work"]

      element.elements.delete_all("dc:title")
      element.elements.delete_all("dc:description")

      title = REXML::Element.new "dc:title"
      title.add_text "#{@iconname}"
      element.add_element title

      description = REXML::Element.new "dc:description"
      description.add_text "#{@context}"
      element.add_element description

      f = File.open(@filename,"w")
      @svg.write(f, -1, false)
      f.close

      puts "New icon #{@filename} created."
      self.new(@filename, :skip_check => true)
    end

    # Create multiple icons with set name and context using selected template
    # into one SVG file
    #
    # filename::
    #  name of the SVG file to be created
    #
    # iconnames::
    #  names of the icons, preferably following freedesktop.org
    #  standards, to be created
    #
    # context::
    #  context of the icon, see THEME_CONTEXTS constant for more info
    #
    # template::
    #  select custom template, either template name (without .svg
    #  extension) that is located in your XDG User Dirs Templates dir or in
    #  /usr/share/icon-artist/templates/#{template}.svg, or a filename of a
    #  custom template (with the .svg extension)
    def self.create_multiple(filename, iconnames, context, template="default", dir = "./")
      if (filename == nil) || (iconnames == nil) ||(iconnames.empty?)
        raise "Nothing to create!"
      end

      if iconnames.length == 1
        return self.create iconnames[0], context, template, dir
      end

      userdir_templates = Icon.get_templatedir
      @template = File.expand_path("#{template}")
      if !File.exist?(@template)
        @template = File.expand_path("#{userdir_templates}/#{template}.svg")
        if !File.exist?(@template)
          @template = "#{ARTIST_SHAREDIR}/templates/#{template}.svg"
          if !File.exist?(@template)
            raise "Could not find either '#{File.expand_path(template)}', '#{File.expand_path(userdir_templates)}/#{template}.svg' or '#{ARTIST_SHAREDIR}/templates/#{template}.svg'."
          end
        end
      end

      @iconname = File.basename(filename, ".svg")
      @context=context
      @filename=File.expand_path(dir)+"/#{filename}"
      @multiple_icons = true
      @icons = iconnames

      if File.exist?(@filename)
        raise "Selected icon (#{@filename}) already exists. \nPlease set another icon name."
      end

      if !(THEME_CONTEXTS.include?(@context))
        out = "Context must be one of: "
        THEME_CONTEXTS.each {|el| out = out + "#{el}, "}
        out = (out + "\n").gsub(", \n", ".\n")
        raise out
      end

      @svg = REXML::Document.new(File.new("#{@template}", 'r'))

      @svg.root.attributes.delete_all("inkscape:export-filename")
      @svg.root.attributes.delete_all("sodipodi:docname")

      element = @svg.root.elements["metadata/rdf:RDF/cc:Work"]

      element.elements.delete_all("dc:title")
      element.elements.delete_all("dc:description")

      title = REXML::Element.new "dc:title"
      title.add_text "#{@iconname}"
      element.add_element title

      description = REXML::Element.new "dc:description"
      description.add_text "#{@context}"
      element.add_element description

      element = @svg.root.elements["//g[@inkscape:label='artwork']"]
      element.add_attribute("inkscape:label", "artwork-#{@icons[0]}")
      element.add_attribute("id", "artwork-#{@icons[0]}")

      element = @svg.root.elements["/svg"]
      @icons.each do |icon|
        next if icon == @icons[0]
        element.add_element "g", {"inkscape:label" => "artwork-#{icon}",
                                  "inkscape:groupmode" => "layer",
                                  "id" => "artwork-#{icon}"}
      end

      f = File.open(@filename,"w")
      @svg.write(f, -1, false)
      f.close

      puts "New icon #{@filename} created."
      self.new(@filename, :skip_check => true)
    end

    # Check if icon in given file follows one-canvas standards
    #
    # filename::
    #  path to the file with the icon
    def self.check(filename)
      @filename = filename
      @svg = REXML::Document.new(File.new("#{filename}", 'r'))

      @icons = Icon.detect_multiple(@svg)
      @multiple_icons = (@icons != nil)

      puts "Checking document metadata..."
      messages = Icon.check_document_metadata(@svg, @filename, @multiple_icons)
      if messages[:errors].length == 0
        puts "Done."
      else
        puts "Failed. Check above messages for more info."
        return messages
      end

      puts "Checking plate layer..."
      msg = Icon.check_layer_plate(@svg)
      messages[:warnings] += msg[:warnings]
      messages[:errors] += msg[:errors]
      if messages[:errors].length == 0
        puts "Done."
      else
        puts "Failed. Check above messages for more info."
        return messages
      end

      puts "Checking artwork layer..."
      msg = Icon.check_layer_artwork(@svg, @multiple_icons)
      messages[:warnings] += msg[:warnings]
      messages[:errors] += msg[:errors]
      if messages[:errors].length == 0
        puts "Done."
      else
        puts "Failed. Check above messages for more info."
        return messages
      end

      puts "\nCheck complete. No fatal errors found."
      if messages[:warnings].length > 0
        if messages[:warnings].length == 1
          puts "There was #{messages[:warnings].length} warning."
        else
          puts "There were #{messages[:warnings].length} warnings."
        end
      end
      return messages
    end

    # Export the PNGs and SVGs to selected directory using standard icon theme
    # directory structure. Returns an array with export parameters which is hash
    # {"name" => icon name, "context" => icon context, "sizes" => exported sizes}
    #
    # dir::
    #  export directory
    #
    # sizes::
    #  set sizes to export
    def export(dir=".",sizes=THEME_SIZES)
      export_dir = File.expand_path(dir)

      if @multiple_icons
        icons = []
        separate_icons (export_dir) unless @separated
        @icon_files.each_value do |icon_file|
          single_icon = IconArtist::Icon.new(icon_file)
          icons += single_icon.export export_dir
        end
        return icons
      else
        exported_sizes = export_pngs(export_dir,sizes)
        if sizes.include?('scalable')
          exported_sizes += export_scalable(export_dir)
        end
        puts "Icon #{@iconname} exported successfully."
        icon = {
          "name" => @iconname,
          "context" => @context,
          "sizes" => exported_sizes
        }
        return [icon]
      end
    end

    # Add symlinks to the icon
    #
    # symlinks::
    #  array of symlinks to be added
    #
    # iconname::
    #  original iconname, ignored when file contains single icon
    def add_symlinks(symlinks, iconname = nil)
      initialize_symlinks unless @symlinks_initialized

      if (@multiple_icons && (@symlinks[iconname] != nil))
        iconlinks = @symlinks[iconname]
      elsif (@symlinks[@iconname] != nil)
        iconlinks = @symlinks[@iconname]
      else
        puts "Warning: Adding symlink for an unknown icon '#{iconname}'."
        return
      end

      symlinks.each do |symlink|
        if !iconlinks[:default].include?(symlink) & !iconlinks[:optional].include?(symlink)
          iconlinks[:optional] += [symlink]
        end
      end
    end

    # Accessor to symlinks. Returs a hash containing :default and :optional
    # symlinks. :default symlinks are found using icon-naming-utils.
    def symlinks
      initialize_symlinks unless @symlinks_initialized
      return @symlinks
    end

    # Accessor to icon contex
    def context
      @context
    end

    # Accessor to icon name
    def iconname
      @iconname
    end

    # Accessor to icon filename
    def filename
      @filename
    end

    # Accessor to multiple_icons boolean
    def multiple_icons
      @multiple_icons
    end

    # Accessor to included icons
    def icons
      @icons
    end

    # Returns separated icon files if :separate is true and loaded icon contains
    # multiple icons or if the file contains only one icon. In other cases it
    # returns nil.

    # options::
    #    :separate => separate multiple icons into multiple files
    #    :tmpdir => where the separated icons go, defauts to "./.separated"
    def icon_files(opts={})
      if @multiple_icons
        if opts[:separate] && !@separated
          if opts[:tmpdir]
            separate_icons(opts[:tmpdir])
          else
            separate_icons('./.separated')
          end
          return @icon_files
        elsif !@separated
          return nil
        else
          return @icon_files
        end

      else
        return {@iconname => @filename}
      end
    end

    def separate_icons(export_dir)
      if @multiple_icons && (!@separated)
        @icon_files = {}
        @icons.each do |icon|
          icon_svg = REXML::Document.new(File.new("#{@filename}", 'r'))
          element = icon_svg.root.elements['//cc:Work/dc:title']
          element.text = icon
          icon_svg.root.elements["/svg"].each_element_with_attribute(
            "inkscape:groupmode", "layer", 0, "g") do |layer|
            layer_name = "artwork-" + icon
            if (layer.attribute("inkscape:label").value == "plate")
              layer.attributes["style"] = "display:none"
            elsif (layer.attribute("inkscape:label").value == layer_name)
              layer.attributes["inkscape:label"] = "artwork"
              layer.attributes["style"] = "display:inline"
            else
              icon_svg.root.elements["/svg"].delete(layer)
            end
          end
          icon_dir = "#{export_dir}/one-cavas"
          out ="#{icon_dir}/#{icon}.svg"
          FileUtils.mkdir_p(icon_dir) unless File.exist?(icon_dir)

          f = File.open(out,"w")
          icon_svg.write(f, -1, false)
          f.close
          puts "Temporary file #{out} has been created."
          @icon_files[icon] = out
        end
      end
      @separated = true
    end
  end
end
